/*
 *  soapui, copyright (C) 2005 Ole Matzura / eviware.com 
 *
 *  SoapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of the GNU Lesser General Public License as published by the Free Software Foundation; 
 *  either version 2.1 of the License, or (at your option) any later version.
 *
 *  SoapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.support;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.wsdl.Binding;
import javax.wsdl.BindingFault;
import javax.wsdl.BindingOperation;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.apache.xmlbeans.SchemaGlobalElement;
import org.apache.xmlbeans.SchemaType;
import org.apache.xmlbeans.SchemaTypeLoader;
import org.apache.xmlbeans.XmlBeans;
import org.apache.xmlbeans.XmlError;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlLineNumber;
import org.apache.xmlbeans.XmlObject;
import org.apache.xmlbeans.XmlOptions;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.impl.wsdl.teststeps.assertions.AssertionError;

/**
 * Class for validating SOAP requests/responses against their definition and schema, requires that
 * the messages follow basic-profile requirements
 *  
 * @author Ole.Matzura
 */

public class WsdlValidator
{
   private final WsdlContext wsdlContext;
   private final static Logger log = Logger.getLogger( WsdlValidator.class );

	public WsdlValidator( WsdlContext wsdlContext )
   {
		this.wsdlContext = wsdlContext;
	}
	
	public AssertionError [] assertRequest( String request, String operationName )
	{
		List<XmlError> errors = new ArrayList<XmlError>(); 
		try
		{
			validateXml(request, errors);

			if (errors.isEmpty())
			{
				validateSoapEnvelope(request, errors);

				BindingOperation bindingOperation = findBindingOperation(operationName);
				if (bindingOperation == null)
				{
					errors.add(XmlError.forMessage("Missing operation ["
							+ operationName + "] in wsdl definition"));
				}
				else
					validateMessage(request, bindingOperation, WsdlUtils
							.getInputParts(bindingOperation), errors);
			}
		}
      catch( XmlException e )
      {
      	errors.addAll( e.getErrors() );
      }
		catch (Exception e)
		{
			errors.add( XmlError.forMessage( e.getMessage() ));
		}
		
		return convertErrors( errors );
	}

	public void validateXml(String request, List<XmlError> errors )
	{
		try
		{
			XmlOptions xmlOptions = new XmlOptions();
			xmlOptions.setLoadLineNumbers();
			xmlOptions.setErrorListener(errors);
			XmlObject xml = XmlObject.Factory.parse(request, xmlOptions);
		}
      catch( XmlException e )
      {
      	errors.addAll( e.getErrors() );
      }
		catch (Exception e)
		{
			errors.add( XmlError.forMessage( e.getMessage() ));
		}
	}

	private AssertionError[] convertErrors(List<XmlError> errors)
	{
      if( errors.size() > 0 )
      {
         List<AssertionError> response = new ArrayList<AssertionError>();
         for (Iterator<XmlError> i = errors.iterator(); i.hasNext();)
         {
            AssertionError assertionError = new AssertionError(i.next());
            if( !response.contains( assertionError ))
            	response.add( assertionError );
         }
         
         return response.toArray( new AssertionError[response.size()] );
      }
      
      return new AssertionError[0];
	}

	public void validateMessage( String request, BindingOperation bindingOperation, Part [] parts, List<XmlError> errors )
	{
		try
      {
         if( !wsdlContext.hasSchemaTypes() )
         {
            errors.add( XmlError.forMessage( "Missing schema types for message"));
         }
         else
         {
         	if( !WsdlUtils.isOutputSoapEncoded( bindingOperation))
            {
         		XmlOptions xmlOptions = new XmlOptions();
               xmlOptions.setLoadLineNumbers();
               XmlObject xml = XmlObject.Factory.parse( request, xmlOptions );
      			
               XmlObject[] paths = xml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" + 
                     "$this/env:Envelope/env:Body/env:Fault");
               
               if( paths.length > 0 )
               {
                  validateSoapFault( bindingOperation, paths[0], errors );
               }
               else if( WsdlUtils.isRpc( wsdlContext.getDefinition(), bindingOperation ))
               {
                  validateRpcLiteral( bindingOperation, parts, xml, errors );
               }
               else
               {
                  validateDocLiteral( bindingOperation, parts, xml, errors );
               }
            }
         	else errors.add( XmlError.forMessage( "Validation of SOAP-Encoded messages not supported"));
         }
      }
      catch ( XmlException e )
      {
      	errors.addAll( e.getErrors() );
      }
      catch (Exception e)
      {
      	errors.add( XmlError.forMessage( e.getMessage() ));
      }
	}
	
	private BindingOperation findBindingOperation(String operationName) throws Exception
	{
		Map services = wsdlContext.getDefinition().getServices();
		Iterator i = services.keySet().iterator();
		while( i.hasNext() )
		{
			Service service = (Service) wsdlContext.getDefinition().getService( (QName) i.next());
			Map ports = service.getPorts();
			
			Iterator iterator = ports.keySet().iterator();
			while( iterator.hasNext() )
			{
				Port port = (Port) service.getPort( (String) iterator.next() );
				BindingOperation bindingOperation = port.getBinding().getBindingOperation( operationName, null, null );
				if( bindingOperation != null ) return bindingOperation;
			}
		}
		
		Map bindings = wsdlContext.getDefinition().getBindings();
		i = bindings.keySet().iterator();
		while( i.hasNext() )
		{
			Binding binding = (Binding) bindings.get( i.next() );
			BindingOperation bindingOperation = binding.getBindingOperation( operationName, null, null );
			if( bindingOperation != null ) return bindingOperation;
		}
		
		return null;
	}

	public AssertionError [] assertResponse( String response, String operationName ) 
	{
		List<XmlError> errors = new ArrayList<XmlError>(); 
		try
		{
			validateXml(response, errors);

			if (errors.isEmpty())
			{
				validateSoapEnvelope(response, errors);

				BindingOperation bindingOperation = findBindingOperation(operationName);
				if (bindingOperation == null)
				{
					errors.add(XmlError.forMessage("Missing operation ["
							+ operationName + "] in wsdl definition"));
				}
				else
					validateMessage(response, bindingOperation, WsdlUtils
							.getOutputParts(bindingOperation), errors);
			}
		}
      catch ( XmlException e )
      {
      	errors.addAll( e.getErrors() );
      }
      catch (Exception e)
      {
      	errors.add( XmlError.forMessage( e.getMessage() ));
      }
		
		
		return convertErrors( errors );
	}
	
	 private void validateDocLiteral(BindingOperation bindingOperation, Part[] outputParts, XmlObject msgXml, List<XmlError> errors) throws Exception
   {
      if( outputParts.length != 1 )
      {
         errors.add( XmlError.forMessage("DocLiteral message must contain 1 part definition" ));
         return;
      }

      Part part = outputParts[0];
      QName elementName = part.getElementName();
      if( elementName != null )
      {
      	// just check for correct message element, other elements are avoided (should create an error)
         XmlObject[] paths = msgXml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" +
               "declare namespace ns='" + elementName.getNamespaceURI() + "';" +
               "$this/env:Envelope/env:Body/ns:" + elementName.getLocalPart() );
         
         if( paths.length == 1 )
         {
            SchemaGlobalElement elm = wsdlContext.getSchemaTypes().findElement( elementName );
            if( elm != null )
            {
					validateMessageBody(errors, elm.getType(), paths[0]);
            }
            else errors.add( XmlError.forMessage("Missing part type in associated schema") );
         }
         else errors.add( XmlError.forMessage("Missing message part with name [" + elementName + "]" ));
      }
      else if( part.getTypeName() != null )
      {
         QName typeName = part.getTypeName();
         
         XmlObject[] paths = msgXml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" +
               "declare namespace ns='" + typeName.getNamespaceURI() + "';" +
               "$this/env:Envelope/env:Body/ns:" + part.getName() );
         
         if( paths.length == 1 )
         {
            SchemaType type = wsdlContext.getSchemaTypes().findType( typeName );
            if( type != null )
            {
            	validateMessageBody( errors, type, paths[0] );
               //XmlObject obj = paths[0].copy().changeType( type );
               //obj.validate( new XmlOptions().setErrorListener( errors ));
            }
            else errors.add(XmlError.forMessage( "Missing part type in associated schema") );
         }
         else errors.add( XmlError.forMessage("Missing message part with name:type [" + 
               part.getName() + ":" + typeName + "]" ));
      }
   }

	private void validateMessageBody(List<XmlError> errors, SchemaType type, XmlObject msg) throws XmlException
	{
		// need to create new body element of correct type from xml text
		// since we want to retain line-numbers
		XmlObject obj = XmlObject.Factory.parse( msg.copy().changeType( type ).xmlText( new XmlOptions().setSaveOuter()), new XmlOptions().setLoadLineNumbers() );
		obj = obj.changeType( type );
		
		// create internal error list
		List list = new ArrayList();
		obj.validate( new XmlOptions().setErrorListener( list ));
		
		// transfer errors for "real" line numbers
		for( int c = 0; c < list.size(); c++ )
		{
			XmlError error = (XmlError) list.get( c );
		   errors.add( XmlError.forLocation( error.getMessage(), error.getSourceName(), 
		   		getLine( msg ) + error.getLine()-1, error.getColumn(), error.getOffset() ));
		}
	}

   private int getLine(XmlObject object)
	{
   	List list = new ArrayList();
		object.newCursor().getAllBookmarkRefs( list );
		for( int c = 0; c < list.size(); c++ )
		{
			if( list.get( c ) instanceof XmlLineNumber )
			{
				return ((XmlLineNumber)list.get(c)).getLine();
			}
		}
		
		return -1;
	}

	private void validateRpcLiteral(BindingOperation bindingOperation, Part[] outputParts, XmlObject msgXml, List<XmlError> errors ) throws Exception
   {
      if( outputParts.length == 0 )
         return;
      
      // get root element
      XmlObject[] paths = msgXml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" +
            "declare namespace ns='" + wsdlContext.getDefinition().getTargetNamespace() + "';" +
            "$this/env:Envelope/env:Body/ns:" + bindingOperation.getName() );
      
      if( paths.length != 1 )
      {
         errors.add( XmlError.forMessage("Missing message wrapper element [" + 
               wsdlContext.getDefinition().getTargetNamespace() + "@" + bindingOperation.getName() ));
      }  
      else
      {
         XmlObject wrapper = paths[0];
         
         for (int i = 0; i < outputParts.length; i++)
         {
            Part part = outputParts[i];
            XmlObject[] children = wrapper.selectChildren( new QName( wsdlContext.getDefinition().getTargetNamespace(), part.getName() ));
            if( children.length != 1 )
            {
               errors.add( XmlError.forMessage("Missing message part [" + part.getName() + "]" ));
            }
            else
            {
               QName typeName = part.getTypeName();
               SchemaType type = wsdlContext.getSchemaTypes().findType( typeName );
               if( type != null )
               {
               	validateMessageBody( errors, type, children[0]);
               }
               else errors.add( XmlError.forMessage("Missing type in associated schema for part [" + part.getName() + "]" ));
            }
         }
      }
   }

   private void validateSoapFault(BindingOperation bindingOperation, XmlObject msgXml, List<XmlError> errors) throws Exception
   {
      Map faults = bindingOperation.getBindingFaults();
      Iterator<BindingFault> i = faults.values().iterator();
      
      while( i.hasNext() )
      {
         BindingFault bindingFault = i.next();
         String faultName = bindingFault.getName();
      
         Part[] faultParts = WsdlUtils.getFaultParts( bindingOperation, faultName );
         if( faultParts.length != 1 ) 
         {
         	log.info( "Missing fault parts in wsdl for fault [" + faultName + "]" );
         	continue;
         }
         
         Part part = faultParts[0];
         QName elementName = part.getElementName();
         if( elementName != null )
         {
            XmlObject[] paths = msgXml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" +
                  "declare namespace ns='" + elementName.getNamespaceURI() + "';" +
                  "//env:Envelope/env:Body/env:Fault/detail/ns:" + elementName.getLocalPart() );
            
            if( paths.length == 1 )
            {
               SchemaGlobalElement elm = wsdlContext.getSchemaTypes().findElement( elementName );
               if( elm != null )
               {
               	validateMessageBody( errors, elm.getType(), paths[0]);
               	break;
               }
               else errors.add( XmlError.forMessage("Missing fault part type in associated schema") );
            }
            else log.info("Missing fault part in message with name [" + elementName + "]");
         }
         // this is not allowed by Basic Profile.. remove?
         else if( part.getTypeName() != null )
         {
            QName typeName = part.getTypeName();
            
            XmlObject[] paths = msgXml.selectPath( "declare namespace env='http://schemas.xmlsoap.org/soap/envelope/';" +
                  "declare namespace ns='" + typeName.getNamespaceURI() + "';" +
                  "$this/env:Envelope/env:Fault/detail/ns:" + part.getName() );
            
            if( paths.length == 1 )
            {
               SchemaType type = wsdlContext.getSchemaTypes().findType( typeName );
               if( type != null )
               {
               	validateMessageBody( errors, type, paths[0]);
               }
               else errors.add( XmlError.forMessage( "Missing part type in associated schema" ) );
            }
            else log.info("Missing message part with name:type [" + part.getName() + ":" + typeName + "]");
         }
      }
   }

   public void validateSoapEnvelope(String soapMessage, List<XmlError> errors) 
   {
      try
		{
			SchemaTypeLoader schema = XmlBeans
					.loadXsd(new XmlObject[] { XmlObject.Factory
							.parse(SoapUI.class.getResource("/soapEnvelope.xsd")) });

			SchemaType envelopeType = schema.findDocumentType(new QName(
					"http://schemas.xmlsoap.org/soap/envelope/", "Envelope"));
			
			XmlOptions xmlOptions = new XmlOptions();
			xmlOptions.setLoadLineNumbers();
			XmlObject xmlObject = schema.parse(soapMessage, envelopeType,
					xmlOptions);
			xmlOptions.setErrorListener(errors);
			xmlObject.validate(xmlOptions);
		}
      catch ( XmlException e )
      {
      	errors.addAll( e.getErrors() );
      }
      catch (Exception e)
      {
      	errors.add( XmlError.forMessage( e.getMessage() ));
      }
   }
}
